from __future__ import absolute_import

import os
import boto3
import pytest

from moto import mock_secretsmanager

from instana.singletons import tracer
from ...helpers import get_first_span_by_filter

pwd = os.path.dirname(os.path.abspath(__file__))

def setup_method():
    """ Clear all spans before a test run """
    tracer.recorder.clear_spans()


@pytest.fixture(scope='function')
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ['AWS_ACCESS_KEY_ID'] = 'testing'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'testing'
    os.environ['AWS_SECURITY_TOKEN'] = 'testing'
    os.environ['AWS_SESSION_TOKEN'] = 'testing'


@pytest.fixture(scope='function')
def secretsmanager(aws_credentials):
    with mock_secretsmanager():
        yield boto3.client('secretsmanager', region_name='us-east-1')


def test_vanilla_list_secrets(secretsmanager):
    result = secretsmanager.list_secrets(MaxResults=123)
    assert result['SecretList'] == []


def test_get_secret_value(secretsmanager):
    result = None

    secretsmanager.put_secret_value(
        SecretId='Uber_Password',
        SecretBinary=b'password1',
        SecretString='password1',
        VersionStages=[
            'string',
        ]
    )
    
    with tracer.start_active_span('test'):
        result = secretsmanager.get_secret_value(SecretId="Uber_Password")

    assert result['Name'] == 'Uber_Password'

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 2

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "boto3"
    boto_span = get_first_span_by_filter(spans, filter)
    assert(boto_span)

    assert(boto_span.t == test_span.t)
    assert(boto_span.p == test_span.s)

    assert(test_span.ec is None)
    assert(boto_span.ec is None)

    assert boto_span.data['boto3']['op'] == 'GetSecretValue'
    assert boto_span.data['boto3']['ep'] == 'https://secretsmanager.us-east-1.amazonaws.com'
    assert boto_span.data['boto3']['reg'] == 'us-east-1'
    assert 'payload' not in boto_span.data['boto3']
    
    assert boto_span.data['http']['status'] == 200
    assert boto_span.data['http']['method'] == 'POST'
    assert boto_span.data['http']['url'] == 'https://secretsmanager.us-east-1.amazonaws.com:443/GetSecretValue'


from __future__ import absolute_import

import time
import asyncio
import aiohttp
import unittest

import tornado
from tornado.httpclient import AsyncHTTPClient

import tests.apps.tornado_server

from instana.singletons import async_tracer, agent
from ..helpers import testenv, get_first_span_by_name, get_first_span_by_filter


class TestTornadoServer(unittest.TestCase):
    async def fetch(self, session, url, headers=None):
        try:
            async with session.get(url, headers=headers) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    async def post(self, session, url, headers=None):
        try:
            async with session.post(url, headers=headers, data={"hello": "post"}) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    def setUp(self):
        """ Clear all spans before a test run """
        self.recorder = async_tracer.recorder
        self.recorder.clear_spans()

        # New event loop for every test
        # self.loop = tornado.ioloop.IOLoop.current()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        self.http_client = AsyncHTTPClient()

    def tearDown(self):
        self.http_client.close()

    def test_get(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        self.assertIsNotNone(tornado_span)
        self.assertIsNotNone(aiohttp_span)
        self.assertIsNotNone(test_span)

        self.assertIsNone(async_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, tornado_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(tornado_span.p, aiohttp_span.s)

        # Synthetic
        self.assertIsNone(tornado_span.sy)
        self.assertIsNone(aiohttp_span.sy)
        self.assertIsNone(test_span.sy)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(tornado_span.ec)

        self.assertEqual(200, tornado_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", tornado_span.data["http"]["url"])
        self.assertIsNone(tornado_span.data["http"]["params"])
        self.assertEqual("GET", tornado_span.data["http"]["method"])
        self.assertIsNotNone(tornado_span.stack)
        self.assertTrue(type(tornado_span.stack) is list)
        self.assertTrue(len(tornado_span.stack) > 1)

        self.assertEqual(200, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        self.assertTrue("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        self.assertTrue("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], tornado_span.s)
        self.assertTrue("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        self.assertTrue("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_post(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.post(session, testenv["tornado_server"] + "/")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        self.assertIsNotNone(tornado_span)
        self.assertIsNotNone(aiohttp_span)
        self.assertIsNotNone(test_span)

        self.assertIsNone(async_tracer.active_span)

        self.assertEqual("tornado-server", tornado_span.n)
        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual("sdk", test_span.n)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, tornado_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(tornado_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(tornado_span.ec)

        self.assertEqual(200, tornado_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", tornado_span.data["http"]["url"])
        self.assertIsNone(tornado_span.data["http"]["params"])
        self.assertEqual("POST", tornado_span.data["http"]["method"])
        self.assertIsNotNone(tornado_span.stack)
        self.assertTrue(type(tornado_span.stack) is list)
        self.assertTrue(len(tornado_span.stack) > 1)

        self.assertEqual(200, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", aiohttp_span.data["http"]["url"])
        self.assertEqual("POST", aiohttp_span.data["http"]["method"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        self.assertTrue("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        self.assertTrue("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], tornado_span.s)
        self.assertTrue("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        self.assertTrue("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_synthetic_request(self):
        async def test():
            headers = {
                'X-Instana-Synthetic': '1'
            }
            
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/", headers=headers)

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        self.assertTrue(tornado_span.sy)
        self.assertIsNone(aiohttp_span.sy)
        self.assertIsNone(test_span.sy)

    def test_get_301(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/301")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(4, len(spans))

        filter = lambda span: span.n == "tornado-server" and span.data["http"]["status"] == 301
        tornado_301_span = get_first_span_by_filter(spans, filter)
        filter = lambda span: span.n == "tornado-server" and span.data["http"]["status"] == 200
        tornado_span = get_first_span_by_filter(spans, filter)
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        self.assertIsNotNone(tornado_301_span)
        self.assertIsNotNone(tornado_span)
        self.assertIsNotNone(aiohttp_span)
        self.assertIsNotNone(test_span)

        self.assertIsNone(async_tracer.active_span)

        self.assertEqual("tornado-server", tornado_301_span.n)
        self.assertEqual("tornado-server", tornado_span.n)
        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual("sdk", test_span.n)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, tornado_span.t)
        self.assertEqual(traceId, tornado_301_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(tornado_301_span.p, aiohttp_span.s)
        self.assertEqual(tornado_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(tornado_301_span.ec)
        self.assertIsNone(tornado_span.ec)

        self.assertEqual(301, tornado_301_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/301", tornado_301_span.data["http"]["url"])
        self.assertIsNone(tornado_span.data["http"]["params"])
        self.assertEqual("GET", tornado_301_span.data["http"]["method"])
        self.assertIsNotNone(tornado_301_span.stack)
        self.assertTrue(type(tornado_301_span.stack) is list)
        self.assertTrue(len(tornado_301_span.stack) > 1)

        self.assertEqual(200, tornado_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", tornado_span.data["http"]["url"])
        self.assertEqual("GET", tornado_span.data["http"]["method"])
        self.assertIsNotNone(tornado_span.stack)
        self.assertTrue(type(tornado_span.stack) is list)
        self.assertTrue(len(tornado_span.stack) > 1)

        self.assertEqual(200, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/301", aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        self.assertTrue("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        self.assertTrue("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], tornado_span.s)
        self.assertTrue("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        self.assertTrue("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_get_405(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/405")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        self.assertIsNotNone(tornado_span)
        self.assertIsNotNone(aiohttp_span)
        self.assertIsNotNone(test_span)

        self.assertIsNone(async_tracer.active_span)

        self.assertEqual("tornado-server", tornado_span.n)
        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual("sdk", test_span.n)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, tornado_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(tornado_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(tornado_span.ec)

        self.assertEqual(405, tornado_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/405", tornado_span.data["http"]["url"])
        self.assertIsNone(tornado_span.data["http"]["params"])
        self.assertEqual("GET", tornado_span.data["http"]["method"])
        self.assertIsNotNone(tornado_span.stack)
        self.assertTrue(type(tornado_span.stack) is list)
        self.assertTrue(len(tornado_span.stack) > 1)

        self.assertEqual(405, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/405", aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        self.assertTrue("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        self.assertTrue("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], tornado_span.s)
        self.assertTrue("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        self.assertTrue("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_get_500(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/500")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        self.assertIsNotNone(tornado_span)
        self.assertIsNotNone(aiohttp_span)
        self.assertIsNotNone(test_span)

        self.assertIsNone(async_tracer.active_span)

        self.assertEqual("tornado-server", tornado_span.n)
        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual("sdk", test_span.n)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, tornado_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(tornado_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(aiohttp_span.ec, 1)
        self.assertEqual(tornado_span.ec, 1)

        self.assertEqual(500, tornado_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/500", tornado_span.data["http"]["url"])
        self.assertIsNone(tornado_span.data["http"]["params"])
        self.assertEqual("GET", tornado_span.data["http"]["method"])
        self.assertIsNotNone(tornado_span.stack)
        self.assertTrue(type(tornado_span.stack) is list)
        self.assertTrue(len(tornado_span.stack) > 1)

        self.assertEqual(500, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/500", aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertEqual('Internal Server Error', aiohttp_span.data["http"]["error"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        self.assertTrue("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        self.assertTrue("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], tornado_span.s)
        self.assertTrue("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        self.assertTrue("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_get_504(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/504")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        self.assertIsNotNone(tornado_span)
        self.assertIsNotNone(aiohttp_span)
        self.assertIsNotNone(test_span)

        self.assertIsNone(async_tracer.active_span)

        self.assertEqual("tornado-server", tornado_span.n)
        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual("sdk", test_span.n)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, tornado_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(tornado_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(aiohttp_span.ec, 1)
        self.assertEqual(tornado_span.ec, 1)

        self.assertEqual(504, tornado_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/504", tornado_span.data["http"]["url"])
        self.assertIsNone(tornado_span.data["http"]["params"])
        self.assertEqual("GET", tornado_span.data["http"]["method"])
        self.assertIsNotNone(tornado_span.stack)
        self.assertTrue(type(tornado_span.stack) is list)
        self.assertTrue(len(tornado_span.stack) > 1)

        self.assertEqual(504, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/504", aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertEqual('Gateway Timeout', aiohttp_span.data["http"]["error"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        self.assertTrue("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        self.assertTrue("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], tornado_span.s)
        self.assertTrue("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        self.assertTrue("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_get_with_params_to_scrub(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["tornado_server"] + "/?secret=yeah")

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        self.assertIsNone(async_tracer.active_span)

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        self.assertIsNotNone(tornado_span)
        self.assertIsNotNone(aiohttp_span)
        self.assertIsNotNone(test_span)

        self.assertEqual("tornado-server", tornado_span.n)
        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual("sdk", test_span.n)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, tornado_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(tornado_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(tornado_span.ec)

        self.assertEqual(200, tornado_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", tornado_span.data["http"]["url"])
        self.assertEqual("secret=<redacted>", tornado_span.data["http"]["params"])
        self.assertEqual("GET", tornado_span.data["http"]["method"])
        self.assertIsNotNone(tornado_span.stack)
        self.assertTrue(type(tornado_span.stack) is list)
        self.assertTrue(len(tornado_span.stack) > 1)

        self.assertEqual(200, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertEqual("secret=<redacted>", aiohttp_span.data["http"]["params"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        self.assertTrue("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        self.assertTrue("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], tornado_span.s)
        self.assertTrue("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        self.assertTrue("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_custom_header_capture(self):
        async def test():
            with async_tracer.start_active_span('test'):
                async with aiohttp.ClientSession() as session:
                    # Hack together a manual custom headers list
                    agent.options.extra_http_headers = [u'X-Capture-This', u'X-Capture-That']

                    headers = dict()
                    headers['X-Capture-This'] = 'this'
                    headers['X-Capture-That'] = 'that'

                    return await self.fetch(session, testenv["tornado_server"] + "/?secret=iloveyou", headers=headers)

        response = tornado.ioloop.IOLoop.current().run_sync(test)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        tornado_span = get_first_span_by_name(spans, "tornado-server")
        aiohttp_span = get_first_span_by_name(spans, "aiohttp-client")
        test_span = get_first_span_by_name(spans, "sdk")

        self.assertIsNotNone(tornado_span)
        self.assertIsNotNone(aiohttp_span)
        self.assertIsNotNone(test_span)

        self.assertIsNone(async_tracer.active_span)

        self.assertEqual("tornado-server", tornado_span.n)
        self.assertEqual("aiohttp-client", aiohttp_span.n)
        self.assertEqual("sdk", test_span.n)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, aiohttp_span.t)
        self.assertEqual(traceId, tornado_span.t)

        # Parent relationships
        self.assertEqual(aiohttp_span.p, test_span.s)
        self.assertEqual(tornado_span.p, aiohttp_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(aiohttp_span.ec)
        self.assertIsNone(tornado_span.ec)

        self.assertEqual(200, tornado_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", tornado_span.data["http"]["url"])
        self.assertEqual("secret=<redacted>", tornado_span.data["http"]["params"])
        self.assertEqual("GET", tornado_span.data["http"]["method"])
        self.assertIsNotNone(tornado_span.stack)
        self.assertTrue(type(tornado_span.stack) is list)
        self.assertTrue(len(tornado_span.stack) > 1)

        self.assertEqual(200, aiohttp_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", aiohttp_span.data["http"]["url"])
        self.assertEqual("GET", aiohttp_span.data["http"]["method"])
        self.assertEqual("secret=<redacted>", aiohttp_span.data["http"]["params"])
        self.assertIsNotNone(aiohttp_span.stack)
        self.assertTrue(type(aiohttp_span.stack) is list)
        self.assertTrue(len(aiohttp_span.stack) > 1)

        self.assertTrue("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        self.assertTrue("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], tornado_span.s)
        self.assertTrue("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        self.assertTrue("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

        self.assertTrue("http.X-Capture-This" in tornado_span.data["custom"]["tags"])
        self.assertEqual('this', tornado_span.data["custom"]["tags"]['http.X-Capture-This'])
        self.assertTrue("http.X-Capture-That" in tornado_span.data["custom"]["tags"])
        self.assertEqual('that', tornado_span.data["custom"]["tags"]['http.X-Capture-That'])



from __future__ import absolute_import

import json
import wrapt
import inspect

from ..log import logger
from ..singletons import tracer


try:
    import boto3
    from boto3.s3 import inject

    def lambda_inject_context(payload, scope):
        """
        When boto3 lambda client 'Invoke' is called, we want to inject the tracing context.
        boto3/botocore has specific requirements:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/lambda.html#Lambda.Client.invoke
        """
        try:
            invoke_payload = payload.get('Payload', {})

            if not isinstance(invoke_payload, dict):
                invoke_payload = json.loads(invoke_payload)

            tracer.inject(scope.span.context, 'http_headers', invoke_payload)
            payload['Payload'] = json.dumps(invoke_payload)
        except Exception:
            logger.debug("non-fatal lambda_inject_context: ", exc_info=True)


    @wrapt.patch_function_wrapper('botocore.client', 'BaseClient._make_api_call')
    def make_api_call_with_instana(wrapped, instance, arg_list, kwargs):
        # pylint: disable=protected-access
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*arg_list, **kwargs)

        with tracer.start_active_span("boto3", child_of=parent_span) as scope:
            try:
                operation = arg_list[0]
                payload = arg_list[1]

                scope.span.set_tag('op', operation)
                scope.span.set_tag('ep', instance._endpoint.host)
                scope.span.set_tag('reg', instance._client_config.region_name)

                scope.span.set_tag('http.url', instance._endpoint.host + ':443/' + arg_list[0])
                scope.span.set_tag('http.method', 'POST')

                # Don't collect payload for SecretsManager
                if not hasattr(instance, 'get_secret_value'):
                    scope.span.set_tag('payload', payload)

                # Inject context when invoking lambdas
                if 'lambda' in instance._endpoint.host and operation == 'Invoke':
                    lambda_inject_context(payload, scope)


            except Exception as exc:
                logger.debug("make_api_call_with_instana: collect error", exc_info=True)

            try:
                result = wrapped(*arg_list, **kwargs)

                if isinstance(result, dict):
                    http_dict = result.get('ResponseMetadata')
                    if isinstance(http_dict, dict):
                        status = http_dict.get('HTTPStatusCode')
                        if status is not None:
                            scope.span.set_tag('http.status_code', status)

                return result
            except Exception as exc:
                scope.span.mark_as_errored({'error': exc})
                raise

    def s3_inject_method_with_instana(wrapped, instance, arg_list, kwargs):
        fas = inspect.getfullargspec(wrapped)
        fas_args = fas.args
        fas_args.remove('self')

        # pylint: disable=protected-access
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*arg_list, **kwargs)

        with tracer.start_active_span("boto3", child_of=parent_span) as scope:
            try:
                operation = wrapped.__name__
                scope.span.set_tag('op', operation)
                scope.span.set_tag('ep', instance._endpoint.host)
                scope.span.set_tag('reg', instance._client_config.region_name)

                scope.span.set_tag('http.url', instance._endpoint.host + ':443/' + operation)
                scope.span.set_tag('http.method', 'POST')

                index = 1
                payload = {}
                arg_length = len(arg_list)

                for arg_name in fas_args:
                    payload[arg_name] = arg_list[index-1]

                    index += 1
                    if index > arg_length:
                        break

                scope.span.set_tag('payload', payload)
            except Exception as exc:
                logger.debug("s3_inject_method_with_instana: collect error", exc_info=True)

            try:
                return wrapped(*arg_list, **kwargs)
            except Exception as exc:
                scope.span.mark_as_errored({'error': exc})
                raise

    for method in ['upload_file', 'upload_fileobj', 'download_file', 'download_fileobj']:
        wrapt.wrap_function_wrapper('boto3.s3.inject', method, s3_inject_method_with_instana)

    logger.debug("Instrumenting boto3")
except ImportError:
    pass



from __future__ import absolute_import

import os
import boto3
import pytest

from moto import mock_s3

from instana.singletons import tracer
from ...helpers import get_first_span_by_filter

pwd = os.path.dirname(os.path.abspath(__file__))
upload_filename = os.path.abspath(pwd + '/../../data/boto3/test_upload_file.jpg')
download_target_filename = os.path.abspath(pwd + '/../../data/boto3/download_target_file.asdf')

def setup_method():
    """ Clear all spans before a test run """
    tracer.recorder.clear_spans()
    os.remove(download_target_filename)


@pytest.fixture(scope='function')
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ['AWS_ACCESS_KEY_ID'] = 'testing'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'testing'
    os.environ['AWS_SECURITY_TOKEN'] = 'testing'
    os.environ['AWS_SESSION_TOKEN'] = 'testing'


@pytest.fixture(scope='function')
def s3(aws_credentials):
    with mock_s3():
        yield boto3.client('s3', region_name='us-east-1')


def test_vanilla_create_bucket(s3):
    # s3 is a fixture defined above that yields a boto3 s3 client.
    # Feel free to instantiate another boto3 S3 client -- Keep note of the region though.
    s3.create_bucket(Bucket="aws_bucket_name")

    result = s3.list_buckets()
    assert len(result['Buckets']) == 1
    assert result['Buckets'][0]['Name'] == 'aws_bucket_name'


def test_s3_create_bucket(s3):
    result = None
    with tracer.start_active_span('test'):
        result = s3.create_bucket(Bucket="aws_bucket_name")

    result = s3.list_buckets()
    assert len(result['Buckets']) == 1
    assert result['Buckets'][0]['Name'] == 'aws_bucket_name'

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 2

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "boto3"
    boto_span = get_first_span_by_filter(spans, filter)
    assert(boto_span)

    assert(boto_span.t == test_span.t)
    assert(boto_span.p == test_span.s)

    assert(test_span.ec is None)
    assert(boto_span.ec is None)

    assert boto_span.data['boto3']['op'] == 'CreateBucket'
    assert boto_span.data['boto3']['ep'] == 'https://s3.amazonaws.com'
    assert boto_span.data['boto3']['reg'] == 'us-east-1'
    assert boto_span.data['boto3']['payload'] == {'Bucket': 'aws_bucket_name'}
    assert boto_span.data['http']['status'] == 200
    assert boto_span.data['http']['method'] == 'POST'
    assert boto_span.data['http']['url'] == 'https://s3.amazonaws.com:443/CreateBucket'


def test_s3_list_buckets(s3):
    result = None
    with tracer.start_active_span('test'):
        result = s3.list_buckets()

    result = s3.list_buckets()
    assert len(result['Buckets']) == 0
    assert result['ResponseMetadata']['HTTPStatusCode'] is 200

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 2

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "boto3"
    boto_span = get_first_span_by_filter(spans, filter)
    assert(boto_span)

    assert(boto_span.t == test_span.t)
    assert(boto_span.p == test_span.s)

    assert(test_span.ec is None)
    assert(boto_span.ec is None)

    assert boto_span.data['boto3']['op'] == 'ListBuckets'
    assert boto_span.data['boto3']['ep'] == 'https://s3.amazonaws.com'
    assert boto_span.data['boto3']['reg'] == 'us-east-1'
    assert boto_span.data['boto3']['payload'] == {}
    assert boto_span.data['http']['status'] == 200
    assert boto_span.data['http']['method'] == 'POST'
    assert boto_span.data['http']['url'] == 'https://s3.amazonaws.com:443/ListBuckets'

def test_s3_vanilla_upload_file(s3):
    object_name = 'aws_key_name'
    bucket_name = 'aws_bucket_name'

    s3.create_bucket(Bucket=bucket_name)
    result = s3.upload_file(upload_filename, bucket_name, object_name)
    assert result is None

def test_s3_upload_file(s3):
    object_name = 'aws_key_name'
    bucket_name = 'aws_bucket_name'

    s3.create_bucket(Bucket=bucket_name)

    result = None
    with tracer.start_active_span('test'):
        s3.upload_file(upload_filename, bucket_name, object_name)

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 2

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "boto3"
    boto_span = get_first_span_by_filter(spans, filter)
    assert(boto_span)

    assert(boto_span.t == test_span.t)
    assert(boto_span.p == test_span.s)

    assert(test_span.ec is None)
    assert(boto_span.ec is None)

    assert boto_span.data['boto3']['op'] == 'upload_file'
    assert boto_span.data['boto3']['ep'] == 'https://s3.amazonaws.com'
    assert boto_span.data['boto3']['reg'] == 'us-east-1'
    payload = {'Filename': upload_filename, 'Bucket': 'aws_bucket_name', 'Key': 'aws_key_name'}
    assert boto_span.data['boto3']['payload'] == payload
    assert boto_span.data['http']['method'] == 'POST'
    assert boto_span.data['http']['url'] == 'https://s3.amazonaws.com:443/upload_file'

def test_s3_upload_file_obj(s3):
    object_name = 'aws_key_name'
    bucket_name = 'aws_bucket_name'

    s3.create_bucket(Bucket=bucket_name)

    result = None
    with tracer.start_active_span('test'):
        with open(upload_filename, "rb") as fd:
            s3.upload_fileobj(fd, bucket_name, object_name)

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 2

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "boto3"
    boto_span = get_first_span_by_filter(spans, filter)
    assert(boto_span)

    assert(boto_span.t == test_span.t)
    assert(boto_span.p == test_span.s)

    assert(test_span.ec is None)
    assert(boto_span.ec is None)

    assert(boto_span.data['boto3']['op'] == 'upload_fileobj')
    assert(boto_span.data['boto3']['ep'] == 'https://s3.amazonaws.com')
    assert(boto_span.data['boto3']['reg'] == 'us-east-1')
    payload = {'Fileobj': "<_io.BufferedReader name='%s'>" % upload_filename, 'Bucket': 'aws_bucket_name', 'Key': 'aws_key_name'}
    assert boto_span.data['boto3']['payload'] == payload
    assert boto_span.data['http']['method'] == 'POST'
    assert boto_span.data['http']['url'] == 'https://s3.amazonaws.com:443/upload_fileobj'

def test_s3_download_file(s3):
    object_name = 'aws_key_name'
    bucket_name = 'aws_bucket_name'

    s3.create_bucket(Bucket=bucket_name)
    s3.upload_file(upload_filename, bucket_name, object_name)

    result = None
    with tracer.start_active_span('test'):
        s3.download_file(bucket_name, object_name, download_target_filename)

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 2

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "boto3"
    boto_span = get_first_span_by_filter(spans, filter)
    assert(boto_span)

    assert(boto_span.t == test_span.t)
    assert(boto_span.p == test_span.s)

    assert(test_span.ec is None)
    assert(boto_span.ec is None)

    assert(boto_span.data['boto3']['op'] == 'download_file')
    assert(boto_span.data['boto3']['ep'] == 'https://s3.amazonaws.com')
    assert(boto_span.data['boto3']['reg'] == 'us-east-1')
    payload = {'Bucket': 'aws_bucket_name', 'Key': 'aws_key_name', 'Filename': '%s' % download_target_filename}
    assert boto_span.data['boto3']['payload'] == payload
    assert boto_span.data['http']['method'] == 'POST'
    assert boto_span.data['http']['url'] == 'https://s3.amazonaws.com:443/download_file'

def test_s3_download_file_obj(s3):
    object_name = 'aws_key_name'
    bucket_name = 'aws_bucket_name'

    s3.create_bucket(Bucket=bucket_name)
    s3.upload_file(upload_filename, bucket_name, object_name)

    result = None
    with tracer.start_active_span('test'):
        with open(download_target_filename, "wb") as fd:
            s3.download_fileobj(bucket_name, object_name, fd)

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 2

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "boto3"
    boto_span = get_first_span_by_filter(spans, filter)
    assert(boto_span)

    assert(boto_span.t == test_span.t)
    assert(boto_span.p == test_span.s)

    assert(test_span.ec is None)
    assert(boto_span.ec is None)

    assert boto_span.data['boto3']['op'] == 'download_fileobj'
    assert boto_span.data['boto3']['ep'] == 'https://s3.amazonaws.com'
    assert boto_span.data['boto3']['reg'] == 'us-east-1'
    assert boto_span.data['http']['method'] == 'POST'
    assert boto_span.data['http']['url'] == 'https://s3.amazonaws.com:443/download_fileobj'


from __future__ import absolute_import

import os
import boto3
import pytest

from moto import mock_ses

from instana.singletons import tracer
from ...helpers import get_first_span_by_filter

pwd = os.path.dirname(os.path.abspath(__file__))

def setup_method():
    """ Clear all spans before a test run """
    tracer.recorder.clear_spans()


@pytest.fixture(scope='function')
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ['AWS_ACCESS_KEY_ID'] = 'testing'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'testing'
    os.environ['AWS_SECURITY_TOKEN'] = 'testing'
    os.environ['AWS_SESSION_TOKEN'] = 'testing'


@pytest.fixture(scope='function')
def ses(aws_credentials):
    with mock_ses():
        yield boto3.client('ses', region_name='us-east-1')


def test_vanilla_verify_email(ses):
    result = ses.verify_email_identity(EmailAddress='pglombardo+instana299@tuta.io')
    assert result['ResponseMetadata']['HTTPStatusCode'] == 200


def test_verify_email(ses):
    result = None

    with tracer.start_active_span('test'):
        result = ses.verify_email_identity(EmailAddress='pglombardo+instana299@tuta.io')

    assert result['ResponseMetadata']['HTTPStatusCode'] == 200

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 2

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "boto3"
    boto_span = get_first_span_by_filter(spans, filter)
    assert(boto_span)

    assert(boto_span.t == test_span.t)
    assert(boto_span.p == test_span.s)

    assert(test_span.ec is None)
    assert(boto_span.ec is None)

    assert boto_span.data['boto3']['op'] == 'VerifyEmailIdentity'
    assert boto_span.data['boto3']['ep'] == 'https://email.us-east-1.amazonaws.com'
    assert boto_span.data['boto3']['reg'] == 'us-east-1'
    assert boto_span.data['boto3']['payload'] == {'EmailAddress': 'pglombardo+instana299@tuta.io'}
   
    assert boto_span.data['http']['status'] == 200
    assert boto_span.data['http']['method'] == 'POST'
    assert boto_span.data['http']['url'] == 'https://email.us-east-1.amazonaws.com:443/VerifyEmailIdentity'


"""
This module contains the classes that represents spans.

InstanaSpan - the OpenTracing based span used during tracing

When an InstanaSpan is finished, it is converted into either an SDKSpan
or RegisteredSpan depending on type.

BaseSpan: Base class containing the commonalities for the two descendants
  - SDKSpan: Class that represents an SDK type span
  - RegisteredSpan: Class that represents a Registered type span
"""
import six

from basictracer.span import BasicSpan
import opentracing.ext.tags as ot_tags

from .log import logger
from .util import DictionaryOfStan


class InstanaSpan(BasicSpan):
    stack = None
    synthetic = False

    def mark_as_errored(self, tags = None):
        """
        Mark this span as errored.

        @param tags: optional tags to add to the span
        """
        try:
            ec = self.tags.get('ec', 0)
            self.set_tag('ec', ec + 1)

            if tags is not None and isinstance(tags, dict):
                for key in tags:
                    self.set_tag(key, tags[key])
        except Exception:
            logger.debug('span.mark_as_errored', exc_info=True)

    def assure_errored(self):
        """
        Make sure that this span is marked as errored.
        @return: None
        """
        try:
            ec = self.tags.get('ec', None)
            if ec is None or ec == 0:
                self.set_tag('ec', 1)
        except Exception:
            logger.debug('span.assure_errored', exc_info=True)

    def log_exception(self, exc):
        """
        Log an exception onto this span.  This will log pertinent info from the exception and
        assure that this span is marked as errored.

        @param e: the exception to log
        """
        try:
            message = ""
            self.mark_as_errored()

            if hasattr(exc, '__str__') and len(str(exc)) > 0:
                message = str(exc)
            elif hasattr(exc, 'message') and exc.message is not None:
                message = exc.message
            else:
                message = repr(exc)

            if self.operation_name in ['rpc-server', 'rpc-client']:
                self.set_tag('rpc.error', message)
            elif self.operation_name == "mysql":
                self.set_tag('mysql.error', message)
            elif self.operation_name == "postgres":
                self.set_tag('pg.error', message)
            elif self.operation_name in RegisteredSpan.HTTP_SPANS:
                self.set_tag('http.error', message)
            elif self.operation_name in ["celery-client", "celery-worker"]:
                self.set_tag('error', message)
            else:
                self.log_kv({'message': message})
        except Exception:
            logger.debug("span.log_exception", exc_info=True)
            raise

class BaseSpan(object):
    sy = None

    def __str__(self):
        return "BaseSpan(%s)" % self.__dict__.__str__()

    def __repr__(self):
        return self.__dict__.__str__()

    def __init__(self, span, source, service_name, **kwargs):
        # pylint: disable=invalid-name
        self.t = span.context.trace_id
        self.p = span.parent_id
        self.s = span.context.span_id
        self.ts = int(round(span.start_time * 1000))
        self.d = int(round(span.duration * 1000))
        self.f = source
        self.ec = span.tags.pop('ec', None)
        self.data = DictionaryOfStan()

        if span.synthetic:
            self.sy = True

        if span.stack:
            self.stack = span.stack

        self.__dict__.update(kwargs)
    
    def _validate_tags(self, tags):
        """
        This method will loop through a set of tags to validate each key and value.

        :param tags: dict of tags
        :return: dict - a filtered set of tags
        """
        filtered_tags = {}
        for key in tags.keys():
            validated_key, validated_value = self._validate_tag(key, tags[key])
            if validated_key is not None and validated_value is not None:
                filtered_tags[validated_key] = validated_value
        return filtered_tags

    def _validate_tag(self, key, value):
        """
        This method will assure that <key> and <value> are valid to set as a tag.
        If <value> fails the check, an attempt will be made to convert it into
        something useful.

        On check failure, this method will return None values indicating that the tag is
        not valid and could not be converted into something useful

        :param key: The tag key
        :param value: The tag value
        :return: Tuple (key, value)
        """
        validated_key = None
        validated_value = None

        try:
            # Tag keys must be some type of text or string type
            if isinstance(key, (six.text_type, six.string_types)):
                validated_key = key[0:1024] # Max key length of 1024 characters

                if isinstance(value, (bool, float, int, list, dict, six.text_type, six.string_types)):
                    validated_value = value
                else:
                    validated_value = self._convert_tag_value(value)
            else:
                logger.debug("(non-fatal) tag names must be strings. tag discarded for %s", type(key))
        except Exception:
            logger.debug("instana.span._validate_tag: ", exc_info=True)

        return (validated_key, validated_value)

    def _convert_tag_value(self, value):
        final_value = None

        try:
            final_value = repr(value)
        except Exception:
            final_value = "(non-fatal) span.set_tag: values must be one of these types: bool, float, int, list, " \
                            "set, str or alternatively support 'repr'. tag discarded"
            logger.debug(final_value, exc_info=True)
            return None
        return final_value


class SDKSpan(BaseSpan):
    ENTRY_KIND = ["entry", "server", "consumer"]
    EXIT_KIND = ["exit", "client", "producer"]

    def __init__(self, span, source, service_name, **kwargs):
        # pylint: disable=invalid-name
        super(SDKSpan, self).__init__(span, source, service_name, **kwargs)

        span_kind = self.get_span_kind(span)

        self.n = "sdk"
        self.k = span_kind[1]

        if self.k == 1 and service_name is not None:
            self.data["service"] = service_name

        self.data["sdk"]["name"] = span.operation_name
        self.data["sdk"]["type"] = span_kind[0]
        self.data["sdk"]["custom"]["tags"] = self._validate_tags(span.tags)

        if span.logs is not None and len(span.logs) > 0:
            logs = DictionaryOfStan()
            for log in span.logs:
                filtered_key_values = self._validate_tags(log.key_values)
                if len(filtered_key_values.keys()) > 0:
                    logs[repr(log.timestamp)] = filtered_key_values
            self.data["sdk"]["custom"]["logs"] = logs

        if "arguments" in span.tags:
            self.data['sdk']['arguments'] = span.tags["arguments"]

        if "return" in span.tags:
            self.data['sdk']['return'] = span.tags["return"]

        if len(span.context.baggage) > 0:
            self.data["baggage"] = span.context.baggage

    def get_span_kind(self, span):
        """
        Will retrieve the `span.kind` tag and return a tuple containing the appropriate string and integer
        values for the Instana backend

        :param span: The span to search for the `span.kind` tag
        :return: Tuple (String, Int)
        """
        kind = ("intermediate", 3)
        if "span.kind" in span.tags:
            if span.tags["span.kind"] in self.ENTRY_KIND:
                kind = ("entry", 1)
            elif span.tags["span.kind"] in self.EXIT_KIND:
                kind = ("exit", 2)
        return kind


class RegisteredSpan(BaseSpan):
    HTTP_SPANS = ("aiohttp-client", "aiohttp-server", "django", "http", "soap", "tornado-client",
                  "tornado-server", "urllib3", "wsgi")

    EXIT_SPANS = ("aiohttp-client", "boto3", "cassandra", "celery-client", "couchbase", "log", "memcache",
                  "mongo", "mysql", "postgres", "rabbitmq", "redis", "rpc-client", "sqlalchemy",
                  "soap", "tornado-client", "urllib3", "pymongo", "gcs")

    ENTRY_SPANS = ("aiohttp-server", "aws.lambda.entry", "celery-worker", "django", "wsgi", "rabbitmq",
                   "rpc-server", "tornado-server")

    LOCAL_SPANS = ("render")

    def __init__(self, span, source, service_name, **kwargs):
        # pylint: disable=invalid-name
        super(RegisteredSpan, self).__init__(span, source, service_name, **kwargs)
        self.n = span.operation_name

        self.k = 1
        if span.operation_name in self.ENTRY_SPANS:
            # entry
            self._populate_entry_span_data(span)
            self.data["service"] = service_name
        elif span.operation_name in self.EXIT_SPANS:
            self.k = 2 # exit
            self._populate_exit_span_data(span)
        elif span.operation_name in self.LOCAL_SPANS:
            self.k = 3 # intermediate span
            self._populate_local_span_data(span)

        if "rabbitmq" in self.data and self.data["rabbitmq"]["sort"] == "consume":
            self.k = 1  # entry

        # Store any leftover tags in the custom section
        if len(span.tags) > 0:
            self.data["custom"]["tags"] = self._validate_tags(span.tags)

    def _populate_entry_span_data(self, span):
        if span.operation_name in self.HTTP_SPANS:
            self._collect_http_tags(span)

        elif span.operation_name == "aws.lambda.entry":
            self.data["lambda"]["arn"] = span.tags.pop('lambda.arn', "Unknown")
            self.data["lambda"]["alias"] = None
            self.data["lambda"]["runtime"] = "python"
            self.data["lambda"]["functionName"] = span.tags.pop('lambda.name', "Unknown")
            self.data["lambda"]["functionVersion"] = span.tags.pop('lambda.version', "Unknown")
            self.data["lambda"]["trigger"] = span.tags.pop('lambda.trigger', None)
            self.data["lambda"]["error"] = None

            trigger_type = self.data["lambda"]["trigger"]

            if trigger_type in ["aws:api.gateway", "aws:application.load.balancer"]:
                self._collect_http_tags(span)
            elif trigger_type == 'aws:cloudwatch.events':
                self.data["lambda"]["cw"]["events"]["id"] = span.tags.pop('data.lambda.cw.events.id', None)
                self.data["lambda"]["cw"]["events"]["more"] = span.tags.pop('lambda.cw.events.more', False)
                self.data["lambda"]["cw"]["events"]["resources"] = span.tags.pop('lambda.cw.events.resources', None)

            elif trigger_type == 'aws:cloudwatch.logs':
                self.data["lambda"]["cw"]["logs"]["group"] = span.tags.pop('lambda.cw.logs.group', None)
                self.data["lambda"]["cw"]["logs"]["stream"] = span.tags.pop('lambda.cw.logs.stream', None)
                self.data["lambda"]["cw"]["logs"]["more"] = span.tags.pop('lambda.cw.logs.more', None)
                self.data["lambda"]["cw"]["logs"]["events"] = span.tags.pop('lambda.cw.logs.events', None)

            elif trigger_type == 'aws:s3':
                self.data["lambda"]["s3"]["events"] = span.tags.pop('lambda.s3.events', None)
            elif trigger_type == 'aws:sqs':
                self.data["lambda"]["sqs"]["messages"] = span.tags.pop('lambda.sqs.messages', None)

        elif span.operation_name == "celery-worker":
            self.data["celery"]["task"] = span.tags.pop('task', None)
            self.data["celery"]["task_id"] = span.tags.pop('task_id', None)
            self.data["celery"]["scheme"] = span.tags.pop('scheme', None)
            self.data["celery"]["host"] = span.tags.pop('host', None)
            self.data["celery"]["port"] = span.tags.pop('port', None)
            self.data["celery"]["retry-reason"] = span.tags.pop('retry-reason', None)
            self.data["celery"]["error"] = span.tags.pop('error', None)

        elif span.operation_name == "rabbitmq":
            self.data["rabbitmq"]["exchange"] = span.tags.pop('exchange', None)
            self.data["rabbitmq"]["queue"] = span.tags.pop('queue', None)
            self.data["rabbitmq"]["sort"] = span.tags.pop('sort', None)
            self.data["rabbitmq"]["address"] = span.tags.pop('address', None)
            self.data["rabbitmq"]["key"] = span.tags.pop('key', None)

        elif span.operation_name == "rpc-server":
            self.data["rpc"]["flavor"] = span.tags.pop('rpc.flavor', None)
            self.data["rpc"]["host"] = span.tags.pop('rpc.host', None)
            self.data["rpc"]["port"] = span.tags.pop('rpc.port', None)
            self.data["rpc"]["call"] = span.tags.pop('rpc.call', None)
            self.data["rpc"]["call_type"] = span.tags.pop('rpc.call_type', None)
            self.data["rpc"]["params"] = span.tags.pop('rpc.params', None)
            self.data["rpc"]["baggage"] = span.tags.pop('rpc.baggage', None)
            self.data["rpc"]["error"] = span.tags.pop('rpc.error', None)
        else:
            logger.debug("SpanRecorder: Unknown entry span: %s" % span.operation_name)

    def _populate_local_span_data(self, span):
        if span.operation_name == "render":
            self.data["render"]["name"] = span.tags.pop('name', None)
            self.data["render"]["type"] = span.tags.pop('type', None)
            self.data["log"]["message"] = span.tags.pop('message', None)
            self.data["log"]["parameters"] = span.tags.pop('parameters', None)
        else:
            logger.debug("SpanRecorder: Unknown local span: %s" % span.operation_name)

    def _populate_exit_span_data(self, span):
        if span.operation_name in self.HTTP_SPANS:
            self._collect_http_tags(span)

        elif span.operation_name == "boto3":
            # boto3 also sends http tags
            self._collect_http_tags(span)

            for tag in ['op', 'ep', 'reg', 'payload', 'error']:
                value = span.tags.pop(tag, None)
                if value is not None:
                    if tag == 'payload':
                        self.data["boto3"][tag] = self._validate_tags(value)
                    else:
                        self.data["boto3"][tag] = value

        elif span.operation_name == "cassandra":
            self.data["cassandra"]["cluster"] = span.tags.pop('cassandra.cluster', None)
            self.data["cassandra"]["query"] = span.tags.pop('cassandra.query', None)
            self.data["cassandra"]["keyspace"] = span.tags.pop('cassandra.keyspace', None)
            self.data["cassandra"]["fetchSize"] = span.tags.pop('cassandra.fetchSize', None)
            self.data["cassandra"]["achievedConsistency"] = span.tags.pop('cassandra.achievedConsistency', None)
            self.data["cassandra"]["triedHosts"] = span.tags.pop('cassandra.triedHosts', None)
            self.data["cassandra"]["fullyFetched"] = span.tags.pop('cassandra.fullyFetched', None)
            self.data["cassandra"]["error"] = span.tags.pop('cassandra.error', None)

        elif span.operation_name == "celery-client":
            self.data["celery"]["task"] = span.tags.pop('task', None)
            self.data["celery"]["task_id"] = span.tags.pop('task_id', None)
            self.data["celery"]["scheme"] = span.tags.pop('scheme', None)
            self.data["celery"]["host"] = span.tags.pop('host', None)
            self.data["celery"]["port"] = span.tags.pop('port', None)
            self.data["celery"]["error"] = span.tags.pop('error', None)

        elif span.operation_name == "couchbase":
            self.data["couchbase"]["hostname"] = span.tags.pop('couchbase.hostname', None)
            self.data["couchbase"]["bucket"] = span.tags.pop('couchbase.bucket', None)
            self.data["couchbase"]["type"] = span.tags.pop('couchbase.type', None)
            self.data["couchbase"]["error"] = span.tags.pop('couchbase.error', None)
            self.data["couchbase"]["error_type"] = span.tags.pop('couchbase.error_type', None)
            self.data["couchbase"]["sql"] = span.tags.pop('couchbase.sql', None)

        elif span.operation_name == "rabbitmq":
            self.data["rabbitmq"]["exchange"] = span.tags.pop('exchange', None)
            self.data["rabbitmq"]["queue"] = span.tags.pop('queue', None)
            self.data["rabbitmq"]["sort"] = span.tags.pop('sort', None)
            self.data["rabbitmq"]["address"] = span.tags.pop('address', None)
            self.data["rabbitmq"]["key"] = span.tags.pop('key', None)

        elif span.operation_name == "redis":
            self.data["redis"]["connection"] = span.tags.pop('connection', None)
            self.data["redis"]["driver"] = span.tags.pop('driver', None)
            self.data["redis"]["command"] = span.tags.pop('command', None)
            self.data["redis"]["error"] = span.tags.pop('redis.error', None)
            self.data["redis"]["subCommands"] = span.tags.pop('subCommands', None)

        elif span.operation_name == "rpc-client":
            self.data["rpc"]["flavor"] = span.tags.pop('rpc.flavor', None)
            self.data["rpc"]["host"] = span.tags.pop('rpc.host', None)
            self.data["rpc"]["port"] = span.tags.pop('rpc.port', None)
            self.data["rpc"]["call"] = span.tags.pop('rpc.call', None)
            self.data["rpc"]["call_type"] = span.tags.pop('rpc.call_type', None)
            self.data["rpc"]["params"] = span.tags.pop('rpc.params', None)
            self.data["rpc"]["baggage"] = span.tags.pop('rpc.baggage', None)
            self.data["rpc"]["error"] = span.tags.pop('rpc.error', None)

        elif span.operation_name == "sqlalchemy":
            self.data["sqlalchemy"]["sql"] = span.tags.pop('sqlalchemy.sql', None)
            self.data["sqlalchemy"]["eng"] = span.tags.pop('sqlalchemy.eng', None)
            self.data["sqlalchemy"]["url"] = span.tags.pop('sqlalchemy.url', None)
            self.data["sqlalchemy"]["err"] = span.tags.pop('sqlalchemy.err', None)

        elif span.operation_name == "mysql":
            self.data["mysql"]["host"] = span.tags.pop('host', None)
            self.data["mysql"]["port"] = span.tags.pop('port', None)
            self.data["mysql"]["db"] = span.tags.pop(ot_tags.DATABASE_INSTANCE, None)
            self.data["mysql"]["user"] = span.tags.pop(ot_tags.DATABASE_USER, None)
            self.data["mysql"]["stmt"] = span.tags.pop(ot_tags.DATABASE_STATEMENT, None)
            self.data["mysql"]["error"] = span.tags.pop('mysql.error', None)

        elif span.operation_name == "postgres":
            self.data["pg"]["host"] = span.tags.pop('host', None)
            self.data["pg"]["port"] = span.tags.pop('port', None)
            self.data["pg"]["db"] = span.tags.pop(ot_tags.DATABASE_INSTANCE, None)
            self.data["pg"]["user"] = span.tags.pop(ot_tags.DATABASE_USER, None)
            self.data["pg"]["stmt"] = span.tags.pop(ot_tags.DATABASE_STATEMENT, None)
            self.data["pg"]["error"] = span.tags.pop('pg.error', None)

        elif span.operation_name == "mongo":
            service = "%s:%s" % (span.tags.pop('host', None), span.tags.pop('port', None))
            namespace = "%s.%s" % (span.tags.pop('db', "?"), span.tags.pop('collection', "?"))

            self.data["mongo"]["service"] = service
            self.data["mongo"]["namespace"] = namespace
            self.data["mongo"]["command"] = span.tags.pop('command', None)
            self.data["mongo"]["filter"] = span.tags.pop('filter', None)
            self.data["mongo"]["json"] = span.tags.pop('json', None)
            self.data["mongo"]["error"] = span.tags.pop('error', None)

        elif span.operation_name == "gcs":
            self.data["gcs"]["op"] = span.tags.pop('gcs.op')
            self.data["gcs"]["bucket"] = span.tags.pop('gcs.bucket', None)
            self.data["gcs"]["object"] = span.tags.pop('gcs.object', None)
            self.data["gcs"]["entity"] = span.tags.pop('gcs.entity', None)
            self.data["gcs"]["range"] = span.tags.pop('gcs.range', None)
            self.data["gcs"]["sourceBucket"] = span.tags.pop('gcs.sourceBucket', None)
            self.data["gcs"]["sourceObject"] = span.tags.pop('gcs.sourceObject', None)
            self.data["gcs"]["sourceObjects"] = span.tags.pop('gcs.sourceObjects', None)
            self.data["gcs"]["destinationBucket"] = span.tags.pop('gcs.destinationBucket', None)
            self.data["gcs"]["destinationObject"] = span.tags.pop('gcs.destinationObject', None)
            self.data["gcs"]["numberOfOperations"] = span.tags.pop('gcs.numberOfOperations', None)
            self.data["gcs"]["projectId"] = span.tags.pop('gcs.projectId', None)
            self.data["gcs"]["accessId"] = span.tags.pop('gcs.accessId', None)

        elif span.operation_name == "log":
            # use last special key values
            for l in span.logs:
                if "message" in l.key_values:
                    self.data["log"]["message"] = l.key_values.pop("message", None)
                if "parameters" in l.key_values:
                    self.data["log"]["parameters"] = l.key_values.pop("parameters", None)
        else:
            logger.debug("SpanRecorder: Unknown exit span: %s" % span.operation_name)

    def _collect_http_tags(self, span):
        self.data["http"]["host"] = span.tags.pop("http.host", None)
        self.data["http"]["url"] = span.tags.pop(ot_tags.HTTP_URL, None)
        self.data["http"]["path"] = span.tags.pop("http.path", None)
        self.data["http"]["params"] = span.tags.pop('http.params', None)
        self.data["http"]["method"] = span.tags.pop(ot_tags.HTTP_METHOD, None)
        self.data["http"]["status"] = span.tags.pop(ot_tags.HTTP_STATUS_CODE, None)
        self.data["http"]["path_tpl"] = span.tags.pop("http.path_tpl", None)
        self.data["http"]["error"] = span.tags.pop('http.error', None)

        if span.operation_name == "soap":
            self.data["soap"]["action"] = span.tags.pop('soap.action', None)



import json
import os
import random
import re
import sys
import time

from collections import defaultdict
import pkg_resources

try:
    from urllib import parse
except ImportError:
    import urlparse as parse
    import urllib

from .log import logger

if sys.version_info.major == 2:
    string_types = basestring
else:
    string_types = str

_rnd = random.Random()
_current_pid = 0

BAD_ID = "BADCAFFE"  # Bad Caffe

# Simple implementation of a nested dictionary.
DictionaryOfStan = lambda: defaultdict(DictionaryOfStan)


def generate_id():
    """ Generate a 64bit base 16 ID for use as a Span or Trace ID """
    global _current_pid

    pid = os.getpid()
    if _current_pid != pid:
        _current_pid = pid
        _rnd.seed(int(1000000 * time.time()) ^ pid)
    new_id = format(_rnd.randint(0, 18446744073709551615), '02x')

    if len(new_id) < 16:
        new_id = new_id.zfill(16)

    return new_id


def header_to_id(header):
    """
    We can receive headers in the following formats:
      1. unsigned base 16 hex string of variable length
      2. [eventual]

    :param header: the header to analyze, validate and convert (if needed)
    :return: a valid ID to be used internal to the tracer
    """
    if not isinstance(header, string_types):
        return BAD_ID

    try:
        # Test that header is truly a hexadecimal value before we try to convert
        int(header, 16)

        length = len(header)
        if length < 16:
            # Left pad ID with zeros
            header = header.zfill(16)
        elif length > 16:
            # Phase 0: Discard everything but the last 16byte
            header = header[-16:]

        return header
    except ValueError:
        return BAD_ID


def to_json(obj):
    """
    Convert obj to json.  Used mostly to convert the classes in json_span.py until we switch to nested
    dicts (or something better)

    :param obj: the object to serialize to json
    :return:  json string
    """
    try:
        def extractor(o):
            if not hasattr(o, '__dict__'):
                logger.debug("Couldn't serialize non dict type: %s", type(o))
                return {}
            else:
                return {k.lower(): v for k, v in o.__dict__.items() if v is not None}

        return json.dumps(obj, default=extractor, sort_keys=False, separators=(',', ':')).encode()
    except Exception:
        logger.debug("to_json non-fatal encoding issue: ", exc_info=True)

def to_pretty_json(obj):
    """
    Convert obj to pretty json.  Used mostly in logging/debugging.

    :param obj: the object to serialize to json
    :return:  json string
    """
    try:
        def extractor(o):
            if not hasattr(o, '__dict__'):
                logger.debug("Couldn't serialize non dict type: %s", type(o))
                return {}
            else:
                return {k.lower(): v for k, v in o.__dict__.items() if v is not None}

        return json.dumps(obj, default=extractor, sort_keys=True, indent=4, separators=(',', ':'))
    except Exception:
        logger.debug("to_pretty_json non-fatal encoding issue: ", exc_info=True)


def get_proc_cmdline(as_string=False):
    """
    Parse the proc file system for the command line of this process.  If not available, then return a default.
    Return is dependent on the value of `as_string`.  If True, return the full command line as a string,
    otherwise a list.
    """
    name = "python"
    if os.path.isfile("/proc/self/cmdline"):
        with open("/proc/self/cmdline") as cmd:
            name = cmd.read()
    else:
        # Most likely not on a *nix based OS.  Return a default
        if as_string is True:
            return name
        else:
            return [name]

    # /proc/self/command line will have strings with null bytes such as "/usr/bin/python\0-s\0-d\0".  This
    # bit will prep the return value and drop the trailing null byte
    parts = name.split('\0')
    parts.pop()

    if as_string is True:
        parts = " ".join(parts)

    return parts


def package_version():
    """
    Determine the version of this package.

    :return: String representing known version
    """
    version = ""
    try:
        version = pkg_resources.get_distribution('instana').version
    except pkg_resources.DistributionNotFound:
        version = 'unknown'

    return version


def contains_secret(candidate, matcher, kwlist):
    """
    This function will indicate whether <candidate> contains a secret as described here:
    https://www.instana.com/docs/setup_and_manage/host_agent/configuration/#secrets

    :param candidate: string to check
    :param matcher: the matcher to use
    :param kwlist: the list of keywords to match
    :return: boolean
    """
    try:
        if candidate is None or candidate == "INSTANA_AGENT_KEY":
            return False

        if not isinstance(kwlist, list):
            logger.debug("contains_secret: bad keyword list")
            return False

        if matcher == 'equals-ignore-case':
            for keyword in kwlist:
                if candidate.lower() == keyword.lower():
                    return True
        elif matcher == 'equals':
            for keyword in kwlist:
                if candidate == keyword:
                    return True
        elif matcher == 'contains-ignore-case':
            for keyword in kwlist:
                if keyword.lower() in candidate:
                    return True
        elif matcher == 'contains':
            for keyword in kwlist:
                if keyword in candidate:
                    return True
        elif matcher == 'regex':
            for regexp in kwlist:
                if re.match(regexp, candidate):
                    return True
        else:
            logger.debug("contains_secret: unknown matcher")
            return False

    except Exception:
        logger.debug("contains_secret", exc_info=True)


def strip_secrets_from_query(qp, matcher, kwlist):
    """
    This function will scrub the secrets from a query param string based on the passed in matcher and kwlist.

    blah=1&secret=password&valid=true will result in blah=1&secret=<redacted>&valid=true

    You can even pass in path query combinations:

    /signup?blah=1&secret=password&valid=true will result in /signup?blah=1&secret=<redacted>&valid=true

    :param qp: a string representing the query params in URL form (unencoded)
    :param matcher: the matcher to use
    :param kwlist: the list of keywords to match
    :return: a scrubbed query param string
    """
    path = None

    try:
        if qp is None:
            return ''

        if not isinstance(kwlist, list):
            logger.debug("strip_secrets_from_query: bad keyword list")
            return qp

        # If there are no key=values, then just return
        if not '=' in qp:
            return qp

        if '?' in qp:
            path, query = qp.split('?')
        else:
            query = qp

        params = parse.parse_qsl(query, keep_blank_values=True)
        redacted = ['<redacted>']

        if matcher == 'equals-ignore-case':
            for keyword in kwlist:
                for index, kv in enumerate(params):
                    if kv[0].lower() == keyword.lower():
                        params[index] = (kv[0], redacted)
        elif matcher == 'equals':
            for keyword in kwlist:
                for index, kv in enumerate(params):
                    if kv[0] == keyword:
                        params[index] = (kv[0], redacted)
        elif matcher == 'contains-ignore-case':
            for keyword in kwlist:
                for index, kv in enumerate(params):
                    if keyword.lower() in kv[0].lower():
                        params[index] = (kv[0], redacted)
        elif matcher == 'contains':
            for keyword in kwlist:
                for index, kv in enumerate(params):
                    if keyword in kv[0]:
                        params[index] = (kv[0], redacted)
        elif matcher == 'regex':
            for regexp in kwlist:
                for index, kv in enumerate(params):
                    if re.match(regexp, kv[0]):
                        params[index] = (kv[0], redacted)
        else:
            logger.debug("strip_secrets_from_query: unknown matcher")
            return qp

        if sys.version_info < (3, 0):
            result = urllib.urlencode(params, doseq=True)
        else:
            result = parse.urlencode(params, doseq=True)
        query = parse.unquote(result)

        if path:
            query = path + '?' + query

        return query
    except Exception:
        logger.debug("strip_secrets_from_query", exc_info=True)


def sql_sanitizer(sql):
    """
    Removes values from valid SQL statements and returns a stripped version.

    :param sql: The SQL statement to be sanitized
    :return: String - A sanitized SQL statement without values.
    """
    return regexp_sql_values.sub('?', sql)


# Used by sql_sanitizer
regexp_sql_values = re.compile(r"('[\s\S][^']*'|\d*\.\d+|\d+|NULL)")


def get_default_gateway():
    """
    Attempts to read /proc/self/net/route to determine the default gateway in use.

    :return: String - the ip address of the default gateway or None if not found/possible/non-existant
    """
    try:
        hip = None
        # The first line is the header line
        # We look for the line where the Destination is 00000000 - that is the default route
        # The Gateway IP is encoded backwards in hex.
        with open("/proc/self/net/route") as routes:
            for line in routes:
                parts = line.split('\t')
                if parts[1] == '00000000':
                    hip = parts[2]

        if hip is not None and len(hip) == 8:
            # Reverse order, convert hex to int
            return "%i.%i.%i.%i" % (int(hip[6:8], 16), int(hip[4:6], 16), int(hip[2:4], 16), int(hip[0:2], 16))

    except Exception:
        logger.warning("get_default_gateway: ", exc_info=True)


def get_py_source(filename):
    """
    Retrieves and returns the source code for any Python
    files requested by the UI via the host agent

    @param filename [String] The fully qualified path to a file
    """
    response = None
    try:
        if regexp_py.search(filename) is None:
            response = {"error": "Only Python source files are allowed. (*.py)"}
        else:
            pysource = ""
            with open(filename, 'r') as pyfile:
                pysource = pyfile.read()

            response = {"data": pysource}

    except Exception as exc:
        response = {"error": str(exc)}

    return response


# Used by get_py_source
regexp_py = re.compile(r"\.py$")


def every(delay, task, name):
    """
    Executes a task every `delay` seconds

    :param delay: the delay in seconds
    :param task: the method to run.  The method should return False if you want the loop to stop.
    :return: None
    """
    next_time = time.time() + delay

    while True:
        time.sleep(max(0, next_time - time.time()))
        try:
            if task() is False:
                break
        except Exception:
            logger.debug("Problem while executing repetitive task: %s", name, exc_info=True)

        # skip tasks if we are behind schedule:
        next_time += (time.time() - next_time) // delay * delay + delay


def determine_service_name():
    """ This function makes a best effort to name this application process. """

    # One environment variable to rule them all
    if "INSTANA_SERVICE_NAME" in os.environ:
        return os.environ["INSTANA_SERVICE_NAME"]

    # Now best effort in naming this process.  No nice package.json like in Node.js
    # so we do best effort detection here.
    app_name = "python"  # the default name
    basename = None

    try:
        if not hasattr(sys, 'argv'):
            proc_cmdline = get_proc_cmdline(as_string=False)
            return os.path.basename(proc_cmdline[0])

        # Get first argument that is not an CLI option
        for candidate in sys.argv:
            if len(candidate) > 0 and candidate[0] != '-':
                basename = candidate
                break

        # If nothing found, fall back to executable
        if basename is None:
            basename = os.path.basename(sys.executable)
        else:
            # Assure leading paths are stripped
            basename = os.path.basename(basename)

        if basename == "gunicorn":
            if 'setproctitle' in sys.modules:
                # With the setproctitle package, gunicorn renames their processes
                # to pretty things - we use those by default
                # gunicorn: master [djface.wsgi]
                # gunicorn: worker [djface.wsgi]
                app_name = get_proc_cmdline(as_string=True)
            else:
                app_name = basename
        elif "FLASK_APP" in os.environ:
            app_name = os.environ["FLASK_APP"]
        elif "DJANGO_SETTINGS_MODULE" in os.environ:
            app_name = os.environ["DJANGO_SETTINGS_MODULE"].split('.')[0]
        elif basename == '':
            if sys.stdout.isatty():
                app_name = "Interactive Console"
            else:
                # No arguments.  Take executable as app_name
                app_name = os.path.basename(sys.executable)
        else:
            # Last chance.  app_name for "python main.py" would be "main.py" here.
            app_name = basename

        # We should have a good app_name by this point.
        # Last conditional, if uwsgi, then wrap the name
        # with the uwsgi process type
        if basename == "uwsgi":
            # We have an app name by this point.  Now if running under
            # uwsgi, augment the app name
            try:
                import uwsgi

                if app_name == "uwsgi":
                    app_name = ""
                else:
                    app_name = " [%s]" % app_name

                if os.getpid() == uwsgi.masterpid():
                    uwsgi_type = "uWSGI master%s"
                else:
                    uwsgi_type = "uWSGI worker%s"

                app_name = uwsgi_type % app_name
            except ImportError:
                pass
    except Exception:
        logger.debug("non-fatal get_application_name: ", exc_info=True)
    finally:
        return app_name


def normalize_aws_lambda_arn(context):
    """
    Parse the AWS Lambda context object for a fully qualified AWS Lambda function ARN.

    This method will ensure that the returned value matches the following ARN pattern:
      arn:aws:lambda:${region}:${account-id}:function:${name}:${version}

    @param context:  AWS Lambda context object
    @return:
    """
    try:
        arn = context.invoked_function_arn
        parts = arn.split(':')

        count = len(parts)
        if count == 7:
            # need to append version
            arn = arn + ':' + context.function_version
        elif count != 8:
            logger.debug("Unexpected ARN parse issue: %s", arn)

        return arn
    except:
        logger.debug("normalize_arn: ", exc_info=True)


def validate_url(url):
    """
    Validate if <url> is a valid url

    Examples:
    - "http://localhost:5000" - valid
    - "http://localhost:5000/path" - valid
    - "sandwich" - invalid

    @param url: string
    @return: Boolean
    """
    try:
        result = parse.urlparse(url)
        return all([result.scheme, result.netloc])
    except:
        return False



from __future__ import absolute_import

import opentracing as ot

from .log import logger
from .util import header_to_id
from .span_context import SpanContext


class BinaryPropagator():
    """
    A Propagator for TEXT_MAP.
    """
    HEADER_KEY_T = b'x-instana-t'
    HEADER_KEY_S = b'x-instana-s'
    HEADER_KEY_L = b'x-instana-l'

    def inject(self, span_context, carrier):
        try:
            trace_id = str.encode(span_context.trace_id)
            span_id = str.encode(span_context.span_id)
            level = str.encode("1")

            if isinstance(carrier, dict) or hasattr(carrier, "__dict__"):
                carrier[self.HEADER_KEY_T] = trace_id
                carrier[self.HEADER_KEY_S] = span_id
                carrier[self.HEADER_KEY_L] = level
            elif isinstance(carrier, list):
                carrier.append((self.HEADER_KEY_T, trace_id))
                carrier.append((self.HEADER_KEY_S, span_id))
                carrier.append((self.HEADER_KEY_L, level))
            elif isinstance(carrier, tuple):
                carrier = carrier.__add__(((self.HEADER_KEY_T, trace_id),))
                carrier = carrier.__add__(((self.HEADER_KEY_S, span_id),))
                carrier = carrier.__add__(((self.HEADER_KEY_L, level),))
            elif hasattr(carrier, '__setitem__'):
                carrier.__setitem__(self.HEADER_KEY_T, trace_id)
                carrier.__setitem__(self.HEADER_KEY_S, span_id)
                carrier.__setitem__(self.HEADER_KEY_L, level)
            else:
                raise Exception("Unsupported carrier type", type(carrier))

            return carrier
        except Exception:
            logger.debug("inject error:", exc_info=True)

    def extract(self, carrier):  # noqa
        trace_id = None
        span_id = None
        level = None

        try:
            if isinstance(carrier, dict) or hasattr(carrier, "__getitem__"):
                dc = carrier
            elif hasattr(carrier, "__dict__"):
                dc = carrier.__dict__
            elif isinstance(carrier, list):
                dc = dict(carrier)
            else:
                raise ot.SpanContextCorruptedException()

            for key, value in dc.items():
                if isinstance(key, str):
                    key = str.encode(key)

                if self.HEADER_KEY_T == key:
                    trace_id = header_to_id(value)
                elif self.HEADER_KEY_S == key:
                    span_id = header_to_id(value)
                elif self.HEADER_KEY_L == key:
                    level = value

            ctx = None
            if trace_id is not None and span_id is not None:
                ctx = SpanContext(span_id=span_id,
                                         trace_id=trace_id,
                                         level=level,
                                         baggage={},
                                         sampled=True)
            return ctx

        except Exception:
            logger.debug("extract error:", exc_info=True)



 #!/usr/bin/env python
# -*- coding: utf-8 -*-
import asyncio
from aiohttp import web

from ...helpers import testenv


testenv["aiohttp_port"] = 10810
testenv["aiohttp_server"] = ("http://127.0.0.1:" + str(testenv["aiohttp_port"]))


def say_hello(request):
    return web.Response(text='Hello, world')


def two_hundred_four(request):
    raise web.HTTPNoContent()


def four_hundred_one(request):
    raise web.HTTPUnauthorized(reason="I must simulate errors.", text="Simulated server error.")


def five_hundred(request):
    return web.HTTPInternalServerError(reason="I must simulate errors.", text="Simulated server error.")


def raise_exception(request):
    raise Exception("Simulated exception")


def aiohttp_server():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    app = web.Application(debug=False)
    app.add_routes([web.get('/', say_hello)])
    app.add_routes([web.get('/204', two_hundred_four)])
    app.add_routes([web.get('/401', four_hundred_one)])
    app.add_routes([web.get('/500', five_hundred)])
    app.add_routes([web.get('/exception', raise_exception)])

    runner = web.AppRunner(app)
    loop.run_until_complete(runner.setup())
    site = web.TCPSite(runner, '127.0.0.1', testenv["aiohttp_port"])

    loop.run_until_complete(site.start())
    loop.run_forever()



from __future__ import absolute_import

import os
import boto3
import pytest
import urllib3

from moto import mock_sqs

import tests.apps.flask_app
from instana.singletons import tracer
from ...helpers import get_first_span_by_filter, testenv


pwd = os.path.dirname(os.path.abspath(__file__))

def setup_method():
    """ Clear all spans before a test run """
    tracer.recorder.clear_spans()


@pytest.fixture(scope='function')
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ['AWS_ACCESS_KEY_ID'] = 'testing'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'testing'
    os.environ['AWS_SECURITY_TOKEN'] = 'testing'
    os.environ['AWS_SESSION_TOKEN'] = 'testing'

@pytest.fixture(scope='function')
def http_client():
    yield urllib3.PoolManager()

@pytest.fixture(scope='function')
def sqs(aws_credentials):
    with mock_sqs():
        yield boto3.client('sqs', region_name='us-east-1')


def test_vanilla_create_queue(sqs):
    result = sqs.create_queue(
    QueueName='SQS_QUEUE_NAME',
    Attributes={
        'DelaySeconds': '60',
        'MessageRetentionPeriod': '86400'
    })
    assert result['ResponseMetadata']['HTTPStatusCode'] == 200


def test_send_message(sqs):
    response = None

    # Create the Queue:
    response = sqs.create_queue(
        QueueName='SQS_QUEUE_NAME',
        Attributes={
            'DelaySeconds': '60',
            'MessageRetentionPeriod': '600'
        }
    )
    assert response['QueueUrl']
    queue_url = response['QueueUrl']

    with tracer.start_active_span('test'):
        response = sqs.send_message(
            QueueUrl=queue_url,
            DelaySeconds=10,
            MessageAttributes={
                'Website': {
                    'DataType': 'String',
                    'StringValue': 'https://www.instana.com'
                },
            },
            MessageBody=('Monitor any application, service, or request '
                         'with Instana Application Performance Monitoring')
        )

    assert response['MessageId']

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 2

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "boto3"
    boto_span = get_first_span_by_filter(spans, filter)
    assert(boto_span)

    assert(boto_span.t == test_span.t)
    assert(boto_span.p == test_span.s)

    assert(test_span.ec is None)
    assert(boto_span.ec is None)

    assert boto_span.data['boto3']['op'] == 'SendMessage'
    assert boto_span.data['boto3']['ep'] == 'https://queue.amazonaws.com'
    assert boto_span.data['boto3']['reg'] == 'us-east-1'

    payload = {'QueueUrl': 'https://queue.amazonaws.com/123456789012/SQS_QUEUE_NAME', 'DelaySeconds': 10, 'MessageAttributes': {'Website': {'DataType': 'String', 'StringValue': 'https://www.instana.com'}}, 'MessageBody': 'Monitor any application, service, or request with Instana Application Performance Monitoring'}
    assert boto_span.data['boto3']['payload'] == payload
    
    assert boto_span.data['http']['status'] == 200
    assert boto_span.data['http']['method'] == 'POST'
    assert boto_span.data['http']['url'] == 'https://queue.amazonaws.com:443/SendMessage'

@mock_sqs
def test_app_boto3_sqs(http_client):
    with tracer.start_active_span('test'):
        response = http_client.request('GET', testenv["wsgi_server"] + '/boto3/sqs')

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 5

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert test_span

    filter = lambda span: span.n == "urllib3"
    http_span = get_first_span_by_filter(spans, filter)
    assert http_span

    filter = lambda span: span.n == "wsgi"
    wsgi_span = get_first_span_by_filter(spans, filter)
    assert wsgi_span

    filter = lambda span: span.n == "boto3" and span.data['boto3']['op'] == 'CreateQueue'
    bcq_span = get_first_span_by_filter(spans, filter)
    assert bcq_span

    filter = lambda span: span.n == "boto3" and span.data['boto3']['op'] == 'SendMessage'
    bsm_span = get_first_span_by_filter(spans, filter)
    assert bsm_span

    assert http_span.t == test_span.t
    assert http_span.p == test_span.s

    assert wsgi_span.t == test_span.t
    assert wsgi_span.p == http_span.s

    assert bcq_span.t == test_span.t
    assert bcq_span.p == wsgi_span.s

    assert bsm_span.t == test_span.t
    assert bsm_span.p == wsgi_span.s




"""
A Collector launches a background thread and continually collects & reports data.  The data
can be any combination of metrics, snapshot data and spans.
"""
import sys
import threading

from ..log import logger
from ..singletons import env_is_test
from ..util import every, DictionaryOfStan


if sys.version_info.major == 2:
    import Queue as queue
else:
    import queue # pylint: disable=import-error


class BaseCollector(object):
    """
    Base class to handle the collection & reporting of snapshot and metric data
    This class launches a background thread to do this work.
    """
    def __init__(self, agent):
        # The agent for this process.  Can be Standard, AWSLambda or Fargate
        self.agent = agent

        # The Queue where we store finished spans before they are sent
        self.span_queue = queue.Queue()

        # The background thread that reports data in a loop every self.report_interval seconds
        self.reporting_thread = None

        # Signal for background thread(s) to shutdown
        self.thread_shutdown = threading.Event()

        # Timestamp in seconds of the last time we sent snapshot data
        self.snapshot_data_last_sent = 0
        # How often to report snapshot data (in seconds)
        self.snapshot_data_interval = 300

        # List of helpers that help out in data collection
        self.helpers = []

        # Lock used syncronize reporting - no updates when sending
        # Used by the background reporting thread.  Used to syncronize report attempts and so
        # that we never have two in progress at once.
        self.background_report_lock = threading.Lock()

        # Reporting interval for the background thread(s)
        self.report_interval = 1

    def start(self):
        """
        Starts the collector and starts reporting as long as the agent is in a ready state.
        @return: None
        """
        if self.agent.can_send():
            logger.debug("BaseCollector.start: launching collection thread")
            self.thread_shutdown.clear()
            self.reporting_thread = threading.Thread(target=self.thread_loop, args=())
            self.reporting_thread.setDaemon(True)
            self.reporting_thread.start()
        else:
            logger.warning("BaseCollector.start: the agent tells us we can't send anything out.")

    def shutdown(self, report_final=True):
        """
        Shuts down the collector and reports any final data.
        @return: None
        """
        logger.debug("Collector.shutdown: Reporting final data.")
        self.thread_shutdown.set()

        if report_final is True:
            self.prepare_and_report_data()

    def thread_loop(self):
        """
        Just a loop that is run in the background thread.
        @return: None
        """
        every(self.report_interval, self.background_report, "Instana Collector: prepare_and_report_data")

    def background_report(self):
        """
        The main work-horse method to report data in the background thread.
        @return: Boolean
        """
        if self.thread_shutdown.is_set():
            logger.debug("Thread shutdown signal is active: Shutting down reporting thread")
            return False
        return self.prepare_and_report_data()

    def should_send_snapshot_data(self):
        """
        Determines if snapshot data should be sent
        @return: Boolean
        """
        logger.debug("BaseCollector: should_send_snapshot_data needs to be overridden")
        return False

    def prepare_payload(self):
        """
        Method to prepare the data to be reported.
        @return: DictionaryOfStan()
        """
        logger.debug("BaseCollector: prepare_payload needs to be overridden")
        return DictionaryOfStan()

    def prepare_and_report_data(self):
        """
        Prepare and report the data payload.
        @return: Boolean
        """
        if env_is_test is True:
            return True

        lock_acquired = self.background_report_lock.acquire(False)
        if lock_acquired:
            try:
                payload = self.prepare_payload()
                self.agent.report_data_payload(payload)
            finally:
                self.background_report_lock.release()
        else:
            logger.debug("prepare_and_report_data: Couldn't acquire lock")
        return True

    def collect_snapshot(self, *argv, **kwargs):
        logger.debug("BaseCollector: collect_snapshot needs to be overridden")

    def queued_spans(self):
        """
        Get all of the queued spans
        @return: list
        """
        spans = []
        while True:
            try:
                span = self.span_queue.get(False)
            except queue.Empty:
                break
            else:
                spans.append(span)
        return spans



"""
Module to handle the work related to the many AWS Lambda Triggers.
"""
import gzip
import json
import base64
from io import BytesIO

from ...log import logger


def get_context(tracer, event):
    # TODO: Search for more types of trigger context
    if is_api_gateway_proxy_trigger(event) or is_application_load_balancer_trigger(event):
        return tracer.extract('http_headers', event['headers'])

    return tracer.extract('http_headers', event)


def is_api_gateway_proxy_trigger(event):
    for key in ["resource", "path", "httpMethod"]:
        if key not in event:
            return False
    return True


def is_application_load_balancer_trigger(event):
    if 'requestContext' in event and 'elb' in event['requestContext']:
        return True
    return False


def is_cloudwatch_trigger(event):
    if "source" in event and 'detail-type' in event:
        if event["source"] == 'aws.events' and event['detail-type'] == 'Scheduled Event':
            return True
    return False


def is_cloudwatch_logs_trigger(event):
    if hasattr(event, 'get') and event.get("awslogs", False) is not False:
        return True
    else:
        return False


def is_s3_trigger(event):
    if "Records" in event:
        if len(event["Records"]) > 0 and event["Records"][0]["eventSource"] == 'aws:s3':
            return True
    return False


def is_sqs_trigger(event):
    if "Records" in event:
        if len(event["Records"]) > 0 and event["Records"][0]["eventSource"] == 'aws:sqs':
            return True
    return False


def read_http_query_params(event):
    """
    Used to parse the Lambda QueryString formats.

    @param event: lambda event dict
    @return: String in the form of "a=b&c=d"
    """
    params = []
    try:
        if event is None or type(event) is not dict:
            return ""

        mvqsp = event.get('multiValueQueryStringParameters', None)
        qsp = event.get('queryStringParameters', None)

        if mvqsp is not None and type(mvqsp) is dict:
            for key in mvqsp:
                params.append("%s=%s" % (key, mvqsp[key]))
            return "&".join(params)
        elif qsp is not None and type(qsp) is dict:
            for key in qsp:
                params.append("%s=%s" % (key, qsp[key]))
            return "&".join(params)
        else:
            return ""
    except:
        logger.debug("read_http_query_params: ", exc_info=True)
        return ""


def capture_extra_headers(event, span, extra_headers):
    """
    Capture the headers specified in `extra_headers` from `event` and log them
    as a tag in the span.

    @param event: the lambda event
    @param span: the lambda entry span
    @param extra_headers: a list of http headers to capture
    @return: None
    """
    try:
        event_headers = event.get("headers", None)

        if event_headers is not None:
            for custom_header in extra_headers:
                for key in event_headers:
                    if key.lower() == custom_header.lower():
                        span.set_tag("http.%s" % custom_header, event_headers[key])
    except:
        logger.debug("capture_extra_headers: ", exc_info=True)


def enrich_lambda_span(agent, span, event, context):
    """
    Extract the required information about this Lambda run (and the trigger) and store the data
    on `span`.

    @param agent: the AWSLambdaAgent in use
    @param span: the Lambda entry span
    @param event: the lambda handler event
    @param context: the lambda handler context
    @return: None
    """
    try:
        span.set_tag('lambda.arn', agent.collector.get_fq_arn())
        span.set_tag('lambda.name', context.function_name)
        span.set_tag('lambda.version', context.function_version)

        if event is None or type(event) is not dict:
            logger.debug("enrich_lambda_span: bad event %s", type(event))
            return

        if is_api_gateway_proxy_trigger(event):
            span.set_tag('lambda.trigger', 'aws:api.gateway')
            span.set_tag('http.method', event["httpMethod"])
            span.set_tag('http.url', event["path"])
            span.set_tag('http.path_tpl', event["resource"])
            span.set_tag('http.params', read_http_query_params(event))

            if agent.options.extra_http_headers is not None:
                capture_extra_headers(event, span, agent.options.extra_http_headers)

        elif is_application_load_balancer_trigger(event):
            span.set_tag('lambda.trigger', 'aws:application.load.balancer')
            span.set_tag('http.method', event["httpMethod"])
            span.set_tag('http.url', event["path"])
            span.set_tag('http.params', read_http_query_params(event))

            if agent.options.extra_http_headers is not None:
                capture_extra_headers(event, span, agent.options.extra_http_headers)

        elif is_cloudwatch_trigger(event):
            span.set_tag('lambda.trigger', 'aws:cloudwatch.events')
            span.set_tag('data.lambda.cw.events.id', event['id'])

            resources = event['resources']
            resource_count = len(event['resources'])
            if resource_count > 3:
                resources = event['resources'][:3]
                span.set_tag('lambda.cw.events.more', True)
            else:
                span.set_tag('lambda.cw.events.more', False)

            report = []
            for item in resources:
                if len(item) > 200:
                    item = item[:200]
                report.append(item)
            span.set_tag('lambda.cw.events.resources', report)

        elif is_cloudwatch_logs_trigger(event):
            span.set_tag('lambda.trigger', 'aws:cloudwatch.logs')

            try:
                if 'awslogs' in event and 'data' in event['awslogs']:
                    data = event['awslogs']['data']
                    decoded_data = base64.b64decode(data)
                    decompressed_data = gzip.GzipFile(fileobj=BytesIO(decoded_data)).read()
                    log_data = json.loads(decompressed_data.decode('utf-8'))

                    span.set_tag('lambda.cw.logs.group', log_data.get('logGroup', None))
                    span.set_tag('lambda.cw.logs.stream', log_data.get('logStream', None))
                    if len(log_data['logEvents']) > 3:
                        span.set_tag('lambda.cw.logs.more', True)
                        events = log_data['logEvents'][:3]
                    else:
                        events = log_data['logEvents']

                    event_data = []
                    for item in events:
                        msg = item.get('message', None)
                        if len(msg) > 200:
                            msg = msg[:200]
                        event_data.append(msg)
                    span.set_tag('lambda.cw.logs.events', event_data)
            except Exception as e:
                span.set_tag('lambda.cw.logs.decodingError', repr(e))
        elif is_s3_trigger(event):
            span.set_tag('lambda.trigger', 'aws:s3')

            if "Records" in event:
                events = []
                for item in event["Records"][:3]:
                    bucket_name = "Unknown"
                    if "s3" in item and "bucket" in item["s3"]:
                        bucket_name = item["s3"]["bucket"]["name"]

                    object_name = ""
                    if "s3" in item and "object" in item["s3"]:
                        object_name = item["s3"]["object"].get("key", "Unknown")

                    if len(object_name) > 200:
                        object_name = object_name[:200]

                    events.append({"event": item['eventName'],
                                   "bucket": bucket_name,
                                   "object": object_name})
                span.set_tag('lambda.s3.events', events)

        elif is_sqs_trigger(event):
            span.set_tag('lambda.trigger', 'aws:sqs')

            if "Records" in event:
                events = []
                for item in event["Records"][:3]:
                    events.append({'queue': item['eventSourceARN']})
                span.set_tag('lambda.sqs.messages', events)
    except:
        logger.debug("enrich_lambda_span: ", exc_info=True)



from __future__ import absolute_import

import unittest

from ..helpers import testenv
from instana.singletons import tracer
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, create_engine


engine = create_engine("postgresql://%s:%s@%s/%s" % (testenv['postgresql_user'], testenv['postgresql_pw'],
                                                     testenv['postgresql_host'], testenv['postgresql_db']))
Base = declarative_base()

class StanUser(Base):
     __tablename__ = 'churchofstan'

     id = Column(Integer, primary_key=True)
     name = Column(String)
     fullname = Column(String)
     password = Column(String)

     def __repr__(self):
        return "<User(name='%s', fullname='%s', password='%s')>" % (
                             self.name, self.fullname, self.password)

Base.metadata.create_all(engine)

stan_user = StanUser(name='IAmStan', fullname='Stan Robot', password='3X}vP66ADoCFT2g?HPvoem2eJh,zWXgd36Rb/{aRq/>7EYy6@EEH4BP(oeXac@mR')

Session = sessionmaker(bind=engine)
Session.configure(bind=engine)

sqlalchemy_url = 'postgresql://%s/%s' % (testenv['postgresql_host'], testenv['postgresql_db'])


class TestSQLAlchemy(unittest.TestCase):
    def setUp(self):
        """ Clear all spans before a test run """
        self.recorder = tracer.recorder
        self.recorder.clear_spans()
        self.session = Session()

    def tearDown(self):
        pass

    def test_session_add(self):
        with tracer.start_active_span('test'):
            self.session.add(stan_user)
            self.session.commit()

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        sql_span = spans[0]
        test_span = spans[1]

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, sql_span.t)

        # Parent relationships
        self.assertEqual(sql_span.p, test_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(sql_span.ec)

        # SQLAlchemy span
        self.assertEqual('sqlalchemy', sql_span.n)
        self.assertFalse('custom' in sql_span.data)
        self.assertTrue('sqlalchemy' in sql_span.data)

        self.assertEqual('postgresql', sql_span.data["sqlalchemy"]["eng"])
        self.assertEqual(sqlalchemy_url, sql_span.data["sqlalchemy"]["url"])
        self.assertEqual('INSERT INTO churchofstan (name, fullname, password) VALUES (%(name)s, %(fullname)s, %(password)s) RETURNING churchofstan.id', sql_span.data["sqlalchemy"]["sql"])
        self.assertIsNone(sql_span.data["sqlalchemy"]["err"])

        self.assertIsNotNone(sql_span.stack)
        self.assertTrue(type(sql_span.stack) is list)
        self.assertGreater(len(sql_span.stack), 0)

    def test_transaction(self):
        result = None
        with tracer.start_active_span('test'):
            with engine.begin() as connection:
                result = connection.execute("select 1")
                result = connection.execute("select (name, fullname, password) from churchofstan where name='doesntexist'")

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        sql_span0 = spans[0]
        sql_span1 = spans[1]
        test_span = spans[2]

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, sql_span0.t)
        self.assertEqual(test_span.t, sql_span1.t)

        # Parent relationships
        self.assertEqual(sql_span0.p, test_span.s)
        self.assertEqual(sql_span1.p, test_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(sql_span0.ec)
        self.assertIsNone(sql_span1.ec)

        # SQLAlchemy span0
        self.assertEqual('sqlalchemy', sql_span0.n)
        self.assertFalse('custom' in sql_span0.data)
        self.assertTrue('sqlalchemy' in sql_span0.data)

        self.assertEqual('postgresql', sql_span0.data["sqlalchemy"]["eng"])
        self.assertEqual(sqlalchemy_url, sql_span0.data["sqlalchemy"]["url"])
        self.assertEqual('select 1', sql_span0.data["sqlalchemy"]["sql"])
        self.assertIsNone(sql_span0.data["sqlalchemy"]["err"])

        self.assertIsNotNone(sql_span0.stack)
        self.assertTrue(type(sql_span0.stack) is list)
        self.assertGreater(len(sql_span0.stack), 0)

        # SQLAlchemy span1
        self.assertEqual('sqlalchemy', sql_span1.n)
        self.assertFalse('custom' in sql_span1.data)
        self.assertTrue('sqlalchemy' in sql_span1.data)

        self.assertEqual('postgresql', sql_span1.data["sqlalchemy"]["eng"])
        self.assertEqual(sqlalchemy_url, sql_span1.data["sqlalchemy"]["url"])
        self.assertEqual("select (name, fullname, password) from churchofstan where name='doesntexist'", sql_span1.data["sqlalchemy"]["sql"])
        self.assertIsNone(sql_span1.data["sqlalchemy"]["err"])

        self.assertIsNotNone(sql_span1.stack)
        self.assertTrue(type(sql_span1.stack) is list)
        self.assertGreater(len(sql_span1.stack), 0)

    def test_error_logging(self):
        with tracer.start_active_span('test'):
            try:
                self.session.execute("htVwGrCwVThisIsInvalidSQLaw4ijXd88")
                self.session.commit()
            except:
                pass

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        sql_span = spans[0]
        test_span = spans[1]

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, sql_span.t)

        # Parent relationships
        self.assertEqual(sql_span.p, test_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIs(sql_span.ec, 1)

        # SQLAlchemy span
        self.assertEqual('sqlalchemy', sql_span.n)

        self.assertFalse('custom' in sql_span.data)
        self.assertTrue('sqlalchemy' in sql_span.data)

        self.assertEqual('postgresql', sql_span.data["sqlalchemy"]["eng"])
        self.assertEqual(sqlalchemy_url, sql_span.data["sqlalchemy"]["url"])
        self.assertEqual('htVwGrCwVThisIsInvalidSQLaw4ijXd88', sql_span.data["sqlalchemy"]["sql"])
        self.assertEqual('syntax error at or near "htVwGrCwVThisIsInvalidSQLaw4ijXd88"\nLINE 1: htVwGrCwVThisIsInvalidSQLaw4ijXd88\n        ^\n', sql_span.data["sqlalchemy"]["err"])

        self.assertIsNotNone(sql_span.stack)
        self.assertTrue(type(sql_span.stack) is list)
        self.assertGreater(len(sql_span.stack), 0)



from __future__ import absolute_import

import os
import boto3
import pytest

from moto import mock_lambda

from instana.singletons import tracer
from ...helpers import get_first_span_by_filter


@pytest.fixture(scope='function')
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ['AWS_ACCESS_KEY_ID'] = 'testing'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'testing'
    os.environ['AWS_SECURITY_TOKEN'] = 'testing'
    os.environ['AWS_SESSION_TOKEN'] = 'testing'


@pytest.fixture(scope='function')
def aws_lambda(aws_credentials):
    with mock_lambda():
        yield boto3.client('lambda', region_name='us-east-1')

def setup_method():
    """ Clear all spans before a test run """
    tracer.recorder.clear_spans()

@pytest.mark.skip("Lambda mocking requires docker")
def test_lambda_invoke(aws_lambda):
    result = None

    with tracer.start_active_span('test'):
        result = aws_lambda.invoke(FunctionName='arn:aws:lambda:us-west-1:410797082306:function:CanaryInACoalMine')

    assert result
    assert len(result['Buckets']) == 1
    assert result['Buckets'][0]['Name'] == 'aws_bucket_name'

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 2

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "boto3"
    boto_span = get_first_span_by_filter(spans, filter)
    assert(boto_span)

    assert(boto_span.t == test_span.t)
    assert(boto_span.p == test_span.s)

    assert(test_span.ec is None)
    assert(boto_span.ec is None)

    assert boto_span.data['boto3']['op'] == 'CreateBucket'
    assert boto_span.data['boto3']['ep'] == 'https://s3.amazonaws.com'
    assert boto_span.data['boto3']['reg'] == 'us-east-1'
    assert boto_span.data['boto3']['payload'] == {'Bucket': 'aws_bucket_name'}
    assert boto_span.data['http']['status'] == 200
    assert boto_span.data['http']['method'] == 'POST'
    assert boto_span.data['http']['url'] == 'https://s3.amazonaws.com:443/CreateBucket'



from __future__ import absolute_import

import time
import asyncio
import unittest

import tornado
from tornado.httpclient import AsyncHTTPClient
from instana.singletons import tornado_tracer

import tests.apps.tornado_server
from ..helpers import testenv

from nose.plugins.skip import SkipTest
raise SkipTest("Non deterministic tests TBR")

class TestTornadoClient(unittest.TestCase):

    def setUp(self):
        """ Clear all spans before a test run """
        self.recorder = tornado_tracer.recorder
        self.recorder.clear_spans()

        # New event loop for every test
        # self.loop = tornado.ioloop.IOLoop.current()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        self.http_client = AsyncHTTPClient()

    def tearDown(self):
        self.http_client.close()

    def test_get(self):
        async def test():
            with tornado_tracer.start_active_span('test'):
                return await self.http_client.fetch(testenv["tornado_server"] + "/")

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))

        server_span = spans[0]
        client_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(tornado_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, client_span.t)
        self.assertEqual(traceId, server_span.t)

        # Parent relationships
        self.assertEqual(client_span.p, test_span.s)
        self.assertEqual(server_span.p, client_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(client_span.ec)
        self.assertIsNone(server_span.ec)

        self.assertEqual("tornado-server", server_span.n)
        self.assertEqual(200, server_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", server_span.data["http"]["url"])
        self.assertIsNone(server_span.data["http"]["params"])
        self.assertEqual("GET", server_span.data["http"]["method"])
        self.assertIsNotNone(server_span.stack)
        self.assertTrue(type(server_span.stack) is list)
        self.assertTrue(len(server_span.stack) > 1)

        self.assertEqual("tornado-client", client_span.n)
        self.assertEqual(200, client_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", client_span.data["http"]["url"])
        self.assertEqual("GET", client_span.data["http"]["method"])
        self.assertIsNotNone(client_span.stack)
        self.assertTrue(type(client_span.stack) is list)
        self.assertTrue(len(client_span.stack) > 1)

        assert("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        assert("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], server_span.s)
        assert("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        assert("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_post(self):
        async def test():
            with tornado_tracer.start_active_span('test'):
                return await self.http_client.fetch(testenv["tornado_server"] + "/", method="POST", body='asdf')

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        server_span = spans[0]
        client_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(tornado_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, client_span.t)
        self.assertEqual(traceId, server_span.t)

        # Parent relationships
        self.assertEqual(client_span.p, test_span.s)
        self.assertEqual(server_span.p, client_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(client_span.ec)
        self.assertIsNone(server_span.ec)

        self.assertEqual("tornado-server", server_span.n)
        self.assertEqual(200, server_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", server_span.data["http"]["url"])
        self.assertIsNone(server_span.data["http"]["params"])
        self.assertEqual("POST", server_span.data["http"]["method"])
        self.assertIsNotNone(server_span.stack)
        self.assertTrue(type(server_span.stack) is list)
        self.assertTrue(len(server_span.stack) > 1)

        self.assertEqual("tornado-client", client_span.n)
        self.assertEqual(200, client_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", client_span.data["http"]["url"])
        self.assertEqual("POST", client_span.data["http"]["method"])
        self.assertIsNotNone(client_span.stack)
        self.assertTrue(type(client_span.stack) is list)
        self.assertTrue(len(client_span.stack) > 1)

        assert("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        assert("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], server_span.s)
        assert("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        assert("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_get_301(self):
        async def test():
            with tornado_tracer.start_active_span('test'):
                return await self.http_client.fetch(testenv["tornado_server"] + "/301")

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(4, len(spans))

        server301_span = spans[0]
        server_span = spans[1]
        client_span = spans[2]
        test_span = spans[3]

        self.assertIsNone(tornado_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, client_span.t)
        self.assertEqual(traceId, server301_span.t)
        self.assertEqual(traceId, server_span.t)

        # Parent relationships
        self.assertEqual(server301_span.p, client_span.s)
        self.assertEqual(client_span.p, test_span.s)
        self.assertEqual(server_span.p, client_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(client_span.ec)
        self.assertIsNone(server_span.ec)

        self.assertEqual("tornado-server", server_span.n)
        self.assertEqual(200, server_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", server_span.data["http"]["url"])
        self.assertIsNone(server_span.data["http"]["params"])
        self.assertEqual("GET", server_span.data["http"]["method"])
        self.assertIsNotNone(server_span.stack)
        self.assertTrue(type(server_span.stack) is list)
        self.assertTrue(len(server_span.stack) > 1)

        self.assertEqual("tornado-server", server301_span.n)
        self.assertEqual(301, server301_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/301", server301_span.data["http"]["url"])
        self.assertIsNone(server301_span.data["http"]["params"])
        self.assertEqual("GET", server301_span.data["http"]["method"])
        self.assertIsNotNone(server301_span.stack)
        self.assertTrue(type(server301_span.stack) is list)
        self.assertTrue(len(server301_span.stack) > 1)

        self.assertEqual("tornado-client", client_span.n)
        self.assertEqual(200, client_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/301", client_span.data["http"]["url"])
        self.assertEqual("GET", client_span.data["http"]["method"])
        self.assertIsNotNone(client_span.stack)
        self.assertTrue(type(client_span.stack) is list)
        self.assertTrue(len(client_span.stack) > 1)

        assert("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        assert("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], server_span.s)
        assert("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        assert("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_get_405(self):
        async def test():
            with tornado_tracer.start_active_span('test'):
                try:
                    return await self.http_client.fetch(testenv["tornado_server"] + "/405")
                except tornado.httpclient.HTTPClientError as e:
                    return e.response

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        server_span = spans[0]
        client_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(tornado_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, client_span.t)
        self.assertEqual(traceId, server_span.t)

        # Parent relationships
        self.assertEqual(client_span.p, test_span.s)
        self.assertEqual(server_span.p, client_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(client_span.ec, 1)
        self.assertIsNone(server_span.ec)

        self.assertEqual("tornado-server", server_span.n)
        self.assertEqual(405, server_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/405", server_span.data["http"]["url"])
        self.assertIsNone(server_span.data["http"]["params"])
        self.assertEqual("GET", server_span.data["http"]["method"])
        self.assertIsNotNone(server_span.stack)
        self.assertTrue(type(server_span.stack) is list)
        self.assertTrue(len(server_span.stack) > 1)

        self.assertEqual("tornado-client", client_span.n)
        self.assertEqual(405, client_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/405", client_span.data["http"]["url"])
        self.assertEqual("GET", client_span.data["http"]["method"])
        self.assertIsNotNone(client_span.stack)
        self.assertTrue(type(client_span.stack) is list)
        self.assertTrue(len(client_span.stack) > 1)

        assert("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        assert("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], server_span.s)
        assert("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        assert("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_get_500(self):
        async def test():
            with tornado_tracer.start_active_span('test'):
                try:
                    return await self.http_client.fetch(testenv["tornado_server"] + "/500")
                except tornado.httpclient.HTTPClientError as e:
                    return e.response

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        server_span = spans[0]
        client_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(tornado_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, client_span.t)
        self.assertEqual(traceId, server_span.t)

        # Parent relationships
        self.assertEqual(client_span.p, test_span.s)
        self.assertEqual(server_span.p, client_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(client_span.ec, 1)
        self.assertEqual(server_span.ec, 1)

        self.assertEqual("tornado-server", server_span.n)
        self.assertEqual(500, server_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/500", server_span.data["http"]["url"])
        self.assertIsNone(server_span.data["http"]["params"])
        self.assertEqual("GET", server_span.data["http"]["method"])
        self.assertIsNotNone(server_span.stack)
        self.assertTrue(type(server_span.stack) is list)
        self.assertTrue(len(server_span.stack) > 1)

        self.assertEqual("tornado-client", client_span.n)
        self.assertEqual(500, client_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/500", client_span.data["http"]["url"])
        self.assertEqual("GET", client_span.data["http"]["method"])
        self.assertIsNotNone(client_span.stack)
        self.assertTrue(type(client_span.stack) is list)
        self.assertTrue(len(client_span.stack) > 1)

        assert("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        assert("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], server_span.s)
        assert("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        assert("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_get_504(self):
        async def test():
            with tornado_tracer.start_active_span('test'):
                try:
                    return await self.http_client.fetch(testenv["tornado_server"] + "/504")
                except tornado.httpclient.HTTPClientError as e:
                    return e.response

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        server_span = spans[0]
        client_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(tornado_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, client_span.t)
        self.assertEqual(traceId, server_span.t)

        # Parent relationships
        self.assertEqual(client_span.p, test_span.s)
        self.assertEqual(server_span.p, client_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(client_span.ec, 1)
        self.assertEqual(server_span.ec, 1)

        self.assertEqual("tornado-server", server_span.n)
        self.assertEqual(504, server_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/504", server_span.data["http"]["url"])
        self.assertIsNone(server_span.data["http"]["params"])
        self.assertEqual("GET", server_span.data["http"]["method"])
        self.assertIsNotNone(server_span.stack)
        self.assertTrue(type(server_span.stack) is list)
        self.assertTrue(len(server_span.stack) > 1)

        self.assertEqual("tornado-client", client_span.n)
        self.assertEqual(504, client_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/504", client_span.data["http"]["url"])
        self.assertEqual("GET", client_span.data["http"]["method"])
        self.assertIsNotNone(client_span.stack)
        self.assertTrue(type(client_span.stack) is list)
        self.assertTrue(len(client_span.stack) > 1)

        assert("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        assert("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], server_span.s)
        assert("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        assert("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)

    def test_get_with_params_to_scrub(self):
        async def test():
            with tornado_tracer.start_active_span('test'):
                return await self.http_client.fetch(testenv["tornado_server"] + "/?secret=yeah")

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        server_span = spans[0]
        client_span = spans[1]
        test_span = spans[2]

        self.assertIsNone(tornado_tracer.active_span)

        # Same traceId
        traceId = test_span.t
        self.assertEqual(traceId, client_span.t)
        self.assertEqual(traceId, server_span.t)

        # Parent relationships
        self.assertEqual(client_span.p, test_span.s)
        self.assertEqual(server_span.p, client_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(client_span.ec)
        self.assertIsNone(server_span.ec)

        self.assertEqual("tornado-server", server_span.n)
        self.assertEqual(200, server_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", server_span.data["http"]["url"])
        self.assertEqual('secret=<redacted>', server_span.data["http"]["params"])
        self.assertEqual("GET", server_span.data["http"]["method"])
        self.assertIsNotNone(server_span.stack)
        self.assertTrue(type(server_span.stack) is list)
        self.assertTrue(len(server_span.stack) > 1)

        self.assertEqual("tornado-client", client_span.n)
        self.assertEqual(200, client_span.data["http"]["status"])
        self.assertEqual(testenv["tornado_server"] + "/", client_span.data["http"]["url"])
        self.assertEqual('secret=<redacted>', client_span.data["http"]["params"])
        self.assertEqual("GET", client_span.data["http"]["method"])
        self.assertIsNotNone(client_span.stack)
        self.assertTrue(type(client_span.stack) is list)
        self.assertTrue(len(client_span.stack) > 1)

        assert("X-Instana-T" in response.headers)
        self.assertEqual(response.headers["X-Instana-T"], traceId)
        assert("X-Instana-S" in response.headers)
        self.assertEqual(response.headers["X-Instana-S"], server_span.s)
        assert("X-Instana-L" in response.headers)
        self.assertEqual(response.headers["X-Instana-L"], '1')
        assert("Server-Timing" in response.headers)
        self.assertEqual(response.headers["Server-Timing"], "intid;desc=%s" % traceId)
